!BOI

! !TITLE:  Ensemble Kalman Filter

! !AUTHORS: Wouter Peters, Andy Jacobson

! !INTRODUCTION:  This is an implementation of the ensemble square-root filter of \citet{whitaker02a}.

!EOI
!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module enkf_common
  !BOP
  ! !MODULE: enkf_common

  ! !DESCRIPTION:
  ! This module manages the (statistical) parameters for the Ensemble
  ! Kalman filter.  It has the information which maps parameters to
  ! fluxes.  It also includes routines for intializing and allocating
  ! variables for the Ensemble Kalman Filter, and for reading and
  ! writing the filter state.
  !
  ! These declarations and routines are needed for both forward
  ! and inverse runs.  The idea is to keep things needed only in
  ! the inverse (or forward) run in separate files.

  !
  ! The current parameter implementation (ct06) makes two assumptions
  ! that we would like to be able to relax in the future:
  !
  ! 1. There is a 1:1 correspondence between parameters and 1x1 grid boxes.
  !
  ! 2. We are scaling net fluxes, not parameters within process models, nor
  !     components of the net flux.
  !
  ! Currently the scaling of fluxes is done in emission\_co2\_all for
  ! the inverse code, and in each flux module for the forward code (in
  ! the emission\_apply subroutine).  This means the parameters must be
  ! made available to the flux modules.  In the future, we can put all
  ! the logic of what a parameter means into the flux modules
  ! themselves.

  ! !USES:
  use GO,                    only : gol, goErr, goPr, goLabel
  use GO,                    only : TDate, NewDate, IncrDate, operator(+), operator(-), rTotal
  use global_data,           only : rcfile
  use toolbox,               only : escape_tm
  use global_data,           only : rcfile
  use Dims,                  only : nlon360, nlat180
  use ParTools,              only : myid, root, maxtrac_per_pe, Par_Barrier, ntracetloc
  use ParTools,              only : proc_tracer, tracer_loc, tracer_active, npes
  use chem_param,            only : ntracet
  use ParTools,              only : nmembers, nmembers_per_pe, nmembersloc
  use io_hdf
  use file_hdf,              only: THdfFile, TSds, Init, Done, WriteAttribute, WriteData, SetDim
  use emission_common,       only: enkf_is_inverse, enkf_is_forward, EmisInputDir
  use enkf_constants

  implicit none
  ! --- const -------------------------

  character(len=*), parameter   ::  mname = 'enkf_common'
  private:: mname

  ! !PUBLIC DATA MEMBERS:
  public

  public:: scalefac
  public:: Init_Parameters, Free_Parameters, Read_Params, Set_Params
  public:: SaveState, savestate_hdf, savestate_sds

  ! made public here but defined in enkf_constants
  public:: nlag, timestep, nparams_in_state,regtype,nreg_land,npar_land,nreg_ocn,npar_ocn

! !LOCAL VARIABLES:
  real, dimension(nlon360,nlat180,maxtrac_per_pe)                 :: scalefac=0.0      ! on 1x1

  character(len=256) :: savestate_filename='savestate.hdf'  ! default value; actual value read from rcfile

  type(THdfFile)            ::  savestate_hdf
  type(TSds)                ::  savestate_sds

  real, dimension(360,180)          :: area_1x1=0.0   ! 1x1 area per gridbox in m^2
  integer, dimension(360,180)       :: fluxregions=0  ! map of regions used in inversion
  real, dimension(nparams_in_state) :: area_state=0.0 ! area per region in state in m^2

  type state_data
     real,dimension(nparams_in_state)                  :: xpc        ! mean prior state 
     real,allocatable,dimension(:,:)  :: pdX        ! prior deviations, non-normalized
  end type state_data

  type(state_data),dimension(nlag),target              :: state_X  ! parameter vector with local set of deviations

  type Q_data
     real,dimension(nparams_in_state,nparams_in_state) :: priorcov ! prior covariance matrix
  end type Q_data
  type(Q_data),target                                  :: state_Q  !

  !EOP


contains

  !BOP
  ! !ROUTINE: Init_Parameters
  !EOP

  !BOC
  subroutine Init_Parameters(status)

    use GO,          only: TrcFile, Init, Done, ReadRc
    use Dims,        only: nregions,idate,itau,itaui,itaue,idatei,idatee,icalendo,iyear0,julday0,kmain,nlon360,nlat180

    integer, intent(out)         ::  status
    character(len=*), parameter  ::  rname = mname//'/Init_Parameters'
    integer                      ::  ilag
    type(TrcFile)                ::  rcF

    do ilag=1,nlag
!       print *, "DEBUG INIT PARMS",ilag,nparams_in_state,nmembers_per_pe
       allocate(state_X(ilag)%pdX(nparams_in_state,nmembers_per_pe))
    enddo

    call ReadRegions
    call CalcAreaState

    if (myid .eq. root) then  ! only root writes to the savestate hdf file.

       call Init( rcF, rcfile, status )
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'start.state', savestate_filename, status )
       IF_NOTOK_RETURN(status=1)
       call Done( rcF, status )
       IF_NOTOK_RETURN(status=1)

       call Init( savestate_hdf, trim(savestate_filename), 'create', status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'idate'      , idate              , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'idatei'     , idatei             , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'idatee'     , idatee             , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'itau'       , itau               , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'itaui'      , itaui              , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'itaue'      , itaue              , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'icalendo'   , icalendo           , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'iyear0'     , iyear0             , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'julday0'    , julday0            , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'nlag'               , nlag                            , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'nmembers'           , nmembers                        , status )
       IF_ERROR_RETURN(status=1)
       call WriteAttribute( savestate_hdf, 'nparams_in_state'  , nparams_in_state                 , status )
       IF_ERROR_RETURN(status=1)

       !WP! By identifying the naming schemes in the attributes of the file we can keep changing it as long as we
       !WP! have a "namingscheme_XXX.rc' file with the proper conventions for each scheme. This file can then be read
       !WP! in our analysis scripts to determine which datasets to look for in savestate.hdf, for example:
       !WP!
       !WP! observed.co2.assimilated     : obs_final_meas  (andy convention)
       !WP! ensemble.posterior.parameter : adX             (release run)
       !WP! stationnames.final           : stationnames_obs_final  (wouter convention)
       !WP!
       !WP! This can keep the naming scheme flexible, will not require us to hardcode these names into all
       !WP! analysis scripts, and will not require each of us to memorize all parameter names.

       !WP!call WriteAttribute( savestate_hdf, 'output_naming_scheme'  , 'wouter_jan06'              , status ) ! release run
       !WP!call WriteAttribute( savestate_hdf, 'output_naming_scheme'  , 'andy_jan07'              , status )  ! andy's tests

       call WriteAttribute( savestate_hdf, 'output_naming_scheme'  , 'wouter_may07'              , status )  ! wouter's proposal
       IF_ERROR_RETURN(status=1)


       call io_write(savestate_hdf%id,360,'lon',180,'lat',area_1x1,'area_1x1',units='m2/box')
       call io_write(savestate_hdf%id,360,'lon',180,'lat',fluxregions,'regions')
       call io_write(savestate_hdf%id,nparams_in_state,'nparams_in_state',area_state,'area',units='m2/ecoregion')

    endif

    status=0

    !EOC

  contains

    !BOP
    ! !ROUTINE: ReadRegions
    !EOP
    subroutine ReadRegions
      !
      ! Subroutine reads in a region definition for inversions on 1x1 degrees. Each
      ! gridbox has a number corresponding to its each region. This map is used
      ! later on for transformations of flux2state and back.
      use io_hdf
      implicit none
      !_____________________IO_________________________________________

      !_____________________locals_____________________________________

      character(len=*), parameter ::  rname = mname//'/ReadRegions'

      !  some stuff for reading routines
      integer                         :: io, sfstart, ifail, sds_id
      integer                         :: n_datasets, n_file_attrs
      integer                         :: istat, attributes, num_type
      integer                         :: sffinfo, sfselect, sfginfo
      integer                         :: sfend, sfrnatt, sfrcatt, sffattr
      integer                         ::  i,j,l,n,m,ind
      character(len=100)              :: filename

      filename=trim(EmisInputDir)//'/transcom_'//trim(regtype)//'.hdf'
      io = sfstart(trim(filename), DFACC_READ)
      if ( io == FAIL ) then 
         call escape_tm('read_regions: Failed to open '//trim(filename))
      endif
      call io_read2d_i16g(io,360,180,fluxregions,'budget_region',ifail)

      if ( ifail /= 0 ) call escape_tm('read_region: Failed to read regions')

    end subroutine ReadRegions

    subroutine CalcAreaState
      !
      ! calc area of each region and each 1x1 gridbox
      ! 
      use geometry,      only: calc_dxy
      use Dims,          only: nlat180,nlon360
      implicit none
      !locals
      character(len=*), parameter ::  rname = mname//'/CalcAreaState'
      integer :: i,j,n,m

      !WP! fill area on 1x1 degree grid and calculate area for each region in the
      !state vector 

      call calc_dxy(area_1x1(1,:),nlat180)

      area_state=0.0

      do j=1,nlat180 
         area_1x1(:,j)=area_1x1(1,j)
         do i=1,nlon360
            n=fluxregions(i,j)
            if(n.gt.nparams_in_state) call escape_tm('declare_kfvar: n exceeds nr of regions')

            area_state(n)=area_state(n)+area_1x1(i,j)

         enddo
      enddo

    end subroutine CalcAreaState

  end subroutine Init_Parameters

  subroutine Free_Parameters(status)

    integer, intent(out)         ::  status
    character(len=*), parameter  ::  rname = mname//'/Free_Parameters'
    integer                      ::  ilag

    do ilag=1,nlag
       deallocate(state_X(ilag)%pdX)
    enddo

    if(myid .eq. root) then
       call Done( savestate_hdf, status )
       IF_NOTOK_RETURN(status=1)
    end if

    status=0

  end subroutine Free_Parameters

  subroutine Read_Params(file_name,status)
    
    use io_hdf
    implicit none

    character(len=*), parameter ::  rname = mname//'/Read_Params'


    integer, intent(out)         ::  status
    character(len=*), intent(in)    :: file_name 

    integer                         :: io, sfstart, ifail, sfend !,n
  
    print *,'[Read_Params] reading ', nparams_in_state,' values from ',trim(file_name)
    io = sfstart(trim(file_name), DFACC_READ)
    if ( io == FAIL ) then
       call escape_tm('Read_Params: Failed to open '//trim(file_name))
    endif
    call io_read1d_32g(io,nparams_in_state,state_X(1)%xpc,'xac_01',status)
    IF_NOTOK_RETURN(status=1)

    !    do n=1,nparams_in_state
    !       print *, '[Read_Params], xac_01(',n,') = ',state_X(1)%xpc(n)
    !    enddo

    ifail=sfend(io)
    if ( io == FAIL ) then
       call escape_tm('Read_Params: Failed to close '//trim(file_name))
    endif
   
    status=0

  end subroutine Read_Params


  subroutine Set_Params(value,status)
    
    implicit none

    character(len=*), parameter  :: rname = mname//'/Set_Params'
    real, intent(in)             :: value
    integer, intent(out)         :: status

    state_X(1)%xpc=value

    status=0

  end subroutine Set_Params

!==================================================================

  subroutine Init_Scalefac(nstate,nomean)
    !
    ! This is the former "State2Flux" subroutine.  It places scaling
    ! factors onto the 1x1 map called scalefac.
    ! 

    implicit none
    ! i/o
    integer,intent(in) :: nstate
    !locals
    integer :: n,mem
    logical,optional :: nomean
    real,dimension(:),pointer :: mean,deviation

    do mem=1,nmembersloc

       mean=> state_X(nstate)%xpc
       deviation=> state_X(nstate)%pdX(:,mem)
       if(present(nomean)) deviation=0.0

       ! central state value
       do n=1,nparams_in_state
          where(fluxregions==n) 
             scalefac(:,:,mem)=mean(n)+deviation(n)
          end where
       enddo

       nullify(mean)
       nullify(deviation)

    enddo ! loop over members

  end subroutine Init_Scalefac


  subroutine SaveState(todo,status)
    !
    ! Subroutine saves each time step in the state to a file
    !
    use GO,         only : TrcFile, Init, Done, ReadRc, goUpCase
    use datetime,    only: tstamp
    use Dims,        only: nregions,idate,itau,itaui,itaue,idatei,idatee,icalendo,iyear0,julday0,kmain,nlon360,nlat180
    use ParTools,    only: ntracetloc
    use chem_param,  only: names
    use user_output_forecast,only: n_fcast, concentration, site_categories, sites, global_R_scaling, obs_rejection_threshold,n_sites_active, evaluate_forecast, n_gph_levs

    implicit none

    ! in/out
    character(len=*),intent(in) :: todo
    integer,intent(out)           :: status

    ! local

    character(len=*), parameter ::  rname = mname//'/SaveState'
    integer                    :: itrace

    !hdf
    INTEGER                    :: sfstart
    INTEGER                    :: sfsnatt
    INTEGER                    :: sfscatt
    INTEGER                    :: io
    INTEGER                    :: istat

    integer           :: region, ind, imr,jmr, lmr,n,m,i
    real,allocatable, dimension(:,:) :: gph_cols

    if(myid.ne.root) return  ! only root writes

    select case(goUpCase(trim(todo)))

    case('FINAL')

       call Init(savestate_sds,savestate_hdf,'itau_obs_final',(/n_fcast/),'int',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'n_final','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%itau,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call WriteAttribute( savestate_hdf, 'n_final', n_fcast, status )
       IF_ERROR_RETURN(status=1)

       call Init(savestate_sds,savestate_hdf,'stationnames_obs_final',(/24,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'station_char','dimensionless',(/(i,i=1,24)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_final','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%ident,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call Init(savestate_sds,savestate_hdf,'eventnumber_obs_final',(/30,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'event_char','dimensionless',(/(i,i=1,30)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_final','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%eventnumber,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call Init(savestate_sds,savestate_hdf,'ccg_flags_obs_final',(/3,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'flags_char','dimensionless',(/(i,i=1,3)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_final','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%ccg_flags,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       ! In inverse mode, concentration%model(1) has the mean concentration.  In forward mode,
       ! this is only the first of five tracers (bg, ff, ocn, bio, fires).
       if (enkf_is_inverse()) then

          ! no call to evalute_forecast; this has already 
          ! been done in enkf_tools' Vectorize() routine.

          call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%model(1),'co2_sim_final',units='ppm')
       endif
       if (enkf_is_forward()) then

          call evaluate_forecast

          call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%model(1),'co2_bg_sim_final',units='ppm') ! background

          concentration%co2_all = concentration%model(1) ! background

          do itrace=2,ntracet ! loop over every *other* component tracer (FF, bio, fires, ocean)
             concentration%co2_component = concentration%model(itrace)-concentration%model(1) ! subtract bg
             concentration%co2_all = concentration%co2_all + concentration%co2_component
             print *,names(itrace)
             call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%co2_component,'co2_'//trim(names(itrace))//'_sim_final',units='ppm')
          end do
          call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%co2_all,'co2_sim_final',units='ppm')
       endif
       call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%obs,'co2_obs_final',units='ppm')   ! obs for timestep itau
       call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%lat,'lat_obs_final',units='deg N')
       call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%lon,'lon_obs_final',units='deg E')
       call io_write(savestate_hdf%id,n_fcast,'n_final',concentration%height,'height_obs_final',units='m above sea level')

       allocate(gph_cols(n_fcast,n_gph_levs))
       do istat=1,n_fcast
          gph_cols(istat,:)=concentration(istat)%gph
       enddo
       call io_write(savestate_hdf%id,n_fcast,'n_final',n_gph_levs,'n_gph_levs',gph_cols,'gph_obs_final', units='m above sea level')
       deallocate(gph_cols)

       return  ! no further tasks

    case('PRIOR')

       call WriteAttribute( savestate_hdf, 'n_fcast'            , n_fcast                          , status )
       IF_ERROR_RETURN(status=1)
       call Init(savestate_sds,savestate_hdf,'stationnames_obs_fcast',(/24,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'station_char','dimensionless',(/(i,i=1,24)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_fcast','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%ident,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call Init(savestate_sds,savestate_hdf,'eventnumber_obs_fcast',(/30,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'event_char','dimensionless',(/(i,i=1,30)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_fcast','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%eventnumber,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call Init(savestate_sds,savestate_hdf,'ccg_flags_obs_fcast',(/3,n_fcast/),'char',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'flags_char','dimensionless',(/(i,i=1,3)/),status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,1,'n_fcast','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%ccg_flags,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)


       call Init(savestate_sds,savestate_hdf,'itau_obs_fcast',(/n_fcast/),'int',status)
       IF_ERROR_RETURN(status=1)
       call SetDim(savestate_sds,0,'n_fcast','dimensionless',(/(i,i=1,n_fcast)/),status)
       IF_ERROR_RETURN(status=1)
       call WriteData(savestate_sds,concentration(:)%itau,status)
       IF_ERROR_RETURN(status=1)
       call Done(savestate_sds,status)
       IF_ERROR_RETURN(status=1)

       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%obs,'co2_obs_fcast',units='ppm')   ! obs for timestep itau

       call evaluate_forecast  ! compute average, convert to ppm
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%model(1),'co2_sim_fcast',units='ppm')
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%lat,'lat_obs_fcast', units='deg N')
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%lon,'lon_obs_fcast',units='deg E')
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%height,'height_obs_fcast', units='m above sea level')

       allocate(gph_cols(n_fcast,n_gph_levs))

       do istat=1,n_fcast
          gph_cols(istat,:)=concentration(istat)%gph
       enddo
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',n_gph_levs,'n_gph_levs',gph_cols,'gph_obs_fcast', units='m above sea level')
       deallocate(gph_cols)

    case('POSTERIOR')

       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%hqhr,'hqhr_sim_fcast', units='ppm^2') ! hqhr 
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%error,'error_sim_fcast', units='ppm') ! model-data mismatch 
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%flag,'flag_sim_fcast') ! obs flag
       call io_write(savestate_hdf%id,n_fcast,'n_fcast',concentration%model(1),'co2_sim_ana',units='ppm')   

    case default
       status=1
       write (gol,*)'Error: illegal label (',trim(todo),')~ in call to SaveState ',status; call goPr
       IF_NOTOK_RETURN(status=1)
    end select

    write (gol,*)'SaveState: sfend returns ',status; call goPr
    IF_NOTOK_RETURN(status=1)

    call tstamp(kmain,itau,'Done writing State')

    !ok
    status=0

  end subroutine SaveState

end module enkf_common
