!### macros #####################################################
!
#define IF_NOTOK_RETURN(action) if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
#define IF_ERROR_RETURN(action) if (status> 0) then; write (gol,'("in ",a)') rname; call goErr; action; return; end if
!
!#################################################################
module user_output_forecast

  use GO        , only : gol, goErr, goPr, goLabel, goUpCase
  use dims,       only : im, jm, lm, dx, xref, dy, yref, xbeg, xend, ybeg, yend
  use dims,       only : nregions, region_name, itaui,itaue, xcyc 
  use chem_param, only : fscale
  use ParTools,   only : myid, root, maxtrac_per_pe, ntracetloc, par_barrier
  use emission_common, only: enkf_is_forward

  implicit none


  character(len=200)           :: ObslistFFilename = ''
  integer                      :: n_fcast
  integer,dimension(6)         :: idate_fcast
  integer                      :: itau_fcast
  integer,parameter            :: window = 4*3600 ! 2 hrs around obs hour

  real         :: global_R_scaling        ! scaling factor for stddev of obs 
  real         :: obs_rejection_threshold ! number of SDs away from model prediction before obs is rejected
  integer      :: n_sites_active          ! number of active sites, dimension of sites array
  integer      :: n_site_categories       ! number of site categories, dimension of site_categories array
  integer      :: n_sites_moved           ! number of sites moved for representivity
  integer,parameter           :: n_gph_levs=15   ! number of model levels (counting up from surface) for which gph is output

  character(len=*), parameter  ::  mname = 'user_output_forecast'

  type observation
     integer,dimension(6)       :: idate     ! date of observation
     character(len=24)           :: ident     ! station identifier
     real                       :: lat       ! latitude of station
     real                       :: lon       ! longitude of station
     real                       :: height    ! height of station
     character(len=30)          :: eventnumber       ! eventnumber
     real,dimension(n_gph_levs)      :: gph       ! the geopotential height of the lowest 15 model levels, averaged like obs

     real                            :: obs=0.0      ! observed value (from file)
     ! In forward mode, model has co2 components of bg, ff, ocean, bio, fires.
     ! In inverse mode, model has ensemble predictions.
     real,dimension(maxtrac_per_pe)  :: model=0.0    ! model predicted background value
     real,dimension(maxtrac_per_pe)  :: residual=0.0 ! obs-(bck+hx_prime)

     ! For forward simulations, need working arrays to compute
     ! individual componenets (background subtracted) and the
     ! sum.  Introduce two reals for this here; they are used in
     ! enkf_common's savestate() routine.
     real                            :: co2_all      
     real                            :: co2_component

     real                            :: flag=0.0
     character(len=3)                :: ccg_flags='...'       ! flag string from the NOAA ccg database
     real                            :: hqhr=0.0     ! value of HQH+R for innovation statistics
     real                            :: error=1000.0 ! model data mismatch == sqrt(R)
     integer                         :: category=0   ! site category code, index into site_category array
     logical                         :: may_localize=.TRUE.
     logical                         :: may_reject=.TRUE.

     integer                    :: itau=0
     integer                    :: n_accumulate=0  ! nr of times concentration was added in get_forecast       

     integer                    :: region 
     integer                    :: is
     integer                    :: js
     integer                    :: lsmin
     integer                    :: lsmax
     integer                    :: dupe=0
  end type observation

  type site_category
     integer		:: code
     character(len=30)  :: name
     real               :: mdm=1000.
     real               :: delta_lat=0.0, delta_lon=0.0
     logical            :: may_localize=.TRUE., may_reject=.TRUE.
  end type site_category

  type(observation), dimension(:), allocatable  :: concentration
  type(site_category), dimension(:), allocatable  :: site_categories
  type(site_category), dimension(:), allocatable  :: sites

  private

  ! private 

  integer,parameter           :: nf_trace =  1
  real                        :: rmf
  ! number of locations to be calculated for 1 model time
  integer,dimension(nf_trace) :: if_trace = (/ 1 /)

  public :: observation,site_category
  public :: read_sites, read_co2_obs, get_forecast, concentration, n_fcast, itau_fcast, idate_fcast, ObslistFfilename
  public :: write_co2_pseudo
  public :: sites, site_categories, n_sites_active, n_site_categories
  public :: global_R_scaling, obs_rejection_threshold
  public :: evaluate_forecast
  public :: n_gph_levs

contains

  ! read the sites and weights information file.  Set up
  subroutine read_sites(status)
    use GO,                  only : TrcFile, Init, Done, ReadRc
    use GO,                  only : TDate, NewDate, IncrDate, operator(+), operator(-), rTotal
    use go_string,           only : goNum2str, goSplitLine, goUpCase
    use toolbox,             only : escape_tm
    use global_data,         only : rcfile

    ! --- in/out -----------------------------

    integer, intent(out)      ::  status

    character(len=500)   :: buffer  ! long because it could include path
    character(len=250)   :: rcval1  
    character(len=250)   :: rcval2  
    type(TrcFile)        :: rcF
    integer              :: i,j
    logical              :: site_found

    character(len=*), parameter ::  rname = mname//'/read_sites'

    if(myid==root) then

       call Init( rcF, rcfile, status )
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'sites.rc', buffer ,status)
       ! print sites.rc file name here (buffer recycled)
       print *, '[read_sites] sites rc file: ' // trim(buffer)
       IF_NOTOK_RETURN(status=1)
       call Done( rcF, status )
       IF_NOTOK_RETURN(status=1)

       call Init( rcF, buffer, status )
       IF_NOTOK_RETURN(status=1)

       call ReadRc( rcF, 'obs.rejection.threshold',obs_rejection_threshold,status)
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'global.R.scaling',global_R_scaling ,status)
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'n.site.categories',n_site_categories,status)
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'n.sites.active', n_sites_active ,status)
       IF_NOTOK_RETURN(status=1)
       call ReadRc( rcF, 'n.sites.moved',n_sites_moved,status)
       IF_NOTOK_RETURN(status=1)

       print *, '[read_sites] obs.rejection.threshold: ',goNum2str(obs_rejection_threshold,'(f5.2)')
       print *, '[read_sites] global_R_scaling: ',goNum2str(global_R_scaling,'(f5.2)')
       print *, '[read_sites] n_site_categories: ', trim(goNum2str(n_site_categories)) 
       print *, '[read_sites] n_sites_active: ',trim(goNum2str(n_sites_active))
       print *, '[read_sites] n_sites_moved: ',trim(goNum2str(n_sites_moved))


       ! Read site categories
       if(allocated(site_categories)) deallocate(site_categories)  ! avoid double allocation
       allocate(site_categories(n_site_categories))
       do i=1, n_site_categories
          call ReadRC( rcF, 'site.category.'//goNum2str(i,'(i3.3)'), buffer,status)
          IF_NOTOK_RETURN(status=1)
          call goSplitLine(buffer, rcval1, ';', rcval2,status)
          IF_NOTOK_RETURN(status=1)
          rcval1 = adjustl(rcval1)
          site_categories(i)%name = goUpCase(trim(rcval1(1:30)))
          site_categories(i)%code = i
          buffer = rcval2
          call goSplitLine(buffer, rcval1, ';', rcval2,status)
          IF_NOTOK_RETURN(status=1)
          read (rcval1,'(f)') site_categories(i)%mdm
          buffer = rcval2
          call goSplitLine(buffer, rcval1, ';', rcval2,status)
          IF_NOTOK_RETURN(status=1)

          if(trim(adjustl(rcval1)).eq.'FALSE') then
             site_categories(i)%may_localize=.FALSE.
          else
             site_categories(i)%may_localize=.TRUE.
          endif

          if(trim(adjustl(rcval2)).eq.'FALSE') then
             site_categories(i)%may_reject=.FALSE.
          else
             site_categories(i)%may_reject=.TRUE.
          endif

          print *, '[read_sites] site category ',trim(goNum2str(i)),' (', & 
               trim(site_categories(i)%name),') has mdm ', & 
               trim(goNum2str(site_categories(i)%mdm)), &
               ' ppm.'
          print *, '       may_localize: ', site_categories(i)%may_localize
          print *, '       may_reject:   ', site_categories(i)%may_reject
       enddo

       ! Read sites
       if(allocated(sites)) deallocate(sites)  ! avoid double allocation
       allocate(sites(n_sites_active))
       do i=1, n_sites_active
          call ReadRC( rcF, 'site.active.'//goNum2str(i,'(i4.4)'), buffer,status)
          IF_NOTOK_RETURN(status=1)
          call goSplitLine(buffer, rcval1, ';', rcval2,status)
          IF_NOTOK_RETURN(status=1)
          rcval1 = goUpCase(trim(adjustl(rcval1)))
          rcval2 = goUpCase(trim(adjustl(rcval2)))
          site_found = .FALSE.
          do j=1, n_site_categories
             if(trim(site_categories(j)%name).eq.rcval2) then
                sites(i)%name = goUpCase(trim(adjustl(rcval1(1:30))))
                sites(i)%code = site_categories(j)%code
                sites(i)%mdm = site_categories(j)%mdm
                sites(i)%may_localize = site_categories(j)%may_localize
                sites(i)%may_reject = site_categories(j)%may_reject
                sites(i)%delta_lat = site_categories(j)%delta_lat
                sites(i)%delta_lon = site_categories(j)%delta_lon
                site_found = .TRUE.
                exit
             endif
          enddo
          if (.not.site_found) then
             call escape_tm('[read_sites] site ' // trim(goNum2str(i)) // ': ' & 
                  // trim(sites(i)%name) // ' with category ' // trim(rcval2) // &
                  ': cannot find matching site category')
          else
             print *, '[read_sites] site ',trim(goNum2str(i)),' ,"', &
                  trim(sites(i)%name),'", category ',trim(rcval2), &
                  ' (code ',trim(goNum2str(sites(i)%code)),')'
          endif
       enddo

       ! Read site moves
       do i=1, n_sites_moved
          call ReadRC( rcF, 'site.move.'//goNum2str(i,'(i3.3)'), buffer,status)
          IF_NOTOK_RETURN(status=1)
          call goSplitLine(buffer, rcval1, ';', rcval2,status)
          IF_NOTOK_RETURN(status=1)
          rcval1 = goUpCase(trim(adjustl(rcval1))) ! the site name
          site_found = .FALSE.
          do j=1, n_sites_active
             if(trim(sites(j)%name).eq.rcval1) then
                buffer=rcval2
                call goSplitLine(buffer,rcval1,';',rcval2,status)
                read (rcval1,'(f)') sites(j)%delta_lat 
                read (rcval2,'(f)') sites(j)%delta_lon
                site_found = .TRUE.
                print *, '[read_sites] site move ',trim(goNum2str(i)),' for "', &
                     trim(sites(j)%name),'": delta-lat=',trim(goNum2str(sites(j)%delta_lat)), &
                     ', delta-lon=',trim(goNum2str(sites(j)%delta_lon))
                exit
             endif
          enddo

          if (.not.site_found) then
             print *, '[read_sites]   Warning!  Site move ', trim(goNum2str(i)), ': '
             print *, '       cannot find site matching name "'// trim(rcval1)//'" in sites array'
          endif
       enddo

       call Done( rcF, status )
       IF_NOTOK_RETURN(status=1)

    endif

  end subroutine read_sites


  subroutine read_co2_obs(status)
    use Meteo,      only : Set
    use Meteo,      only : temper_dat, humid_dat, gph_dat
    use binas,      only : ae, twopi, grav
    use dims,       only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend
    use dims,       only : nregions, region_name, itaur, xcyc
    use chem_param, only : ntrace, ntracet, names, fscale
    use toolbox,    only : escape_tm
    use datetime,   only : date2tau
    use go_string,  only : goUpCase   
#ifdef MPI
    use mpi_const,  only : my_real,mpi_comm_world,ierr,mpi_integer,mpi_character
#endif
    implicit none

    ! local
    integer                           :: slen1 !string length
    integer                           :: slen2 !string length
    integer                           :: slen3 !string length
    character(len=200)                :: dummy
    integer                           :: fstatus 
    integer                           :: i_stat, j_stat
    integer                           :: sunit0=333
    integer                           :: region
    integer,dimension(6)              :: idate_sim
    integer                           :: itau_sim, isim, ispec
    ! x,y resolution (in degrees) for current region 
    real                              :: dxr, dyr 
    character(len=8)                  :: name_spec    !CMK changed from 6-->8
    integer           ::  status
    logical                           :: site_found
    integer                           :: i_dupe, n_dupes
    integer, dimension(:),allocatable :: dupes

    ! start

    call Set(    gph_dat, status, used=.true. )
    call Set(    temper_dat, status, used=.true. )
    call Set(    humid_dat, status, used=.true. )

    if(myid==root) then

       !=======================================================
       ! open StationFile (with list of stations)
       !   "station" means "observation" in this context (arj)
       !=======================================================

       open(UNIT=sunit0, FORM='FORMATTED', &
            FILE='./'//trim(ObslistFFilename) , &
            ERR=1000)      

       ! count number of stations
       n_fcast=0
       stat_loop: DO
          read(sunit0,  '(a)', END=100) dummy
          n_fcast = n_fcast+1 
       end do stat_loop
100    continue
       print *, '_________________________________________________' &
            //'_____________________________'
       print *, ' preparing to read ', n_fcast, ' obs from file: ' // ObslistFFilename
       print *, '_________________________________________________' &
            //'_____________________________'

    endif

#ifdef MPI    
    call MPI_BCAST(n_fcast           ,1,      MPI_INTEGER, root ,MPI_COMM_WORLD,ierr)  
#endif

    if(n_fcast.eq.0) call escape_tm( 'read_co2_obs: no obs in file ' // ObslistFFilename) 

    if(allocated(concentration)) deallocate(concentration)  ! avoid double allocation

    allocate(concentration(n_fcast))



    allocate(dupes(n_fcast))


    dyr = dy/yref(1)
    dxr = dx/xref(1)

    !===============
    ! read stations
    !===============

    if(myid==root) then
       rewind(sunit0)
       do i_stat=1, n_fcast 

          !  Note 26 Jun 07:  The "ident" field generally contains the (Globalview-style) siteid.
          !  The "eventnumber" field is a unique identifier for the sample, usually the
          !  GMD database's event number.  We have some measurements that do not exist
          !  in the GMD database for which Ken and Andy made up a pseudo-eventnumber
          !  (it has site location and time info in it).  

          read(sunit0,  '(a24,1x,f8.2,f8.2,f8.1,1x,6i4,x,f9.3,x,a3,x,a30)') &
               concentration(i_stat)%ident, &
               concentration(i_stat)%lat, &
               concentration(i_stat)%lon, &
               concentration(i_stat)%height, &
               concentration(i_stat)%idate, &
               concentration(i_stat)%obs, &
               concentration(i_stat)%ccg_flags, &
               concentration(i_stat)%eventnumber

          ! goUpCase is *probably* unnecessary
          concentration(i_stat)%ident=goUpCase(trim(adjustl(concentration(i_stat)%ident)))

          call date2tau(concentration(i_stat)%idate,concentration(i_stat)%itau)  ! fill itau

          ! The following code detects observations that fall right
          ! onto the integration start or end time.  For such
          ! observations, the sample time is moved one hour forward or
          ! backward into the middle of the integration period. The
          ! fix causes a different sampling time to be written out,
          ! since the observation's tau value itself is changed.  This
          ! makes output appear to be inconsistent with input obs list
          ! files.  Also, since forward simulations have a different
          ! itaui than inverse simulations, outputs for the same
          ! observation from those two types of simulations can appear
          ! to have different sample times.  These annoyances are not
          ! worth the benefit of increasing the averaging period from
          ! its worst-case two hours to the intended three hours.
          !
          ! -Andy, 20 Aug 2009
          !
          ! if(abs(concentration(i_stat)%itau-itaui).lt.1) concentration(i_stat)%itau=concentration(i_stat)%itau+3600
          ! if(abs(concentration(i_stat)%itau-itaue).lt.1) concentration(i_stat)%itau=concentration(i_stat)%itau-3600



          ! Note that the is and js components for the observation locations are
          ! set *before* the site moves are applied.  Is this a bug?
          !  -Andy, 7 Nov 2006

          concentration(i_stat)%is = &
               int((concentration(i_stat)%lon-float(xbeg(1)))/dxr + 0.99999) 
          concentration(i_stat)%js = &
               int((concentration(i_stat)%lat-float(ybeg(1)))/dyr + 0.99999)
          concentration(i_stat)%lsmin = 999
          concentration(i_stat)%lsmax = 0

          ! Check to see whether this site is in the "sites" array.  
          ! Only sites used in the inversion will actually be present in the sites array;
          ! those sampled but not optimized are not necessarily in the array.
          !
          ! Move sites if required.
          ! 
          ! Set model-data mismatch to 1000ppm for obs not in sites array.

          ! guilty until proven innocent.
          site_found = .FALSE.
          concentration(i_stat)%flag=1
          concentration(i_stat)%may_localize=.TRUE.
          concentration(i_stat)%may_reject=.TRUE.

          do j_stat = 1,n_sites_active

             if(goUpCase(trim(concentration(i_stat)%ident)) .eq. goUpCase(trim(sites(j_stat)%name))) then

                site_found = .TRUE.

                concentration(i_stat)%flag=0
                concentration(i_stat)%error=global_R_scaling*sites(j_stat)%mdm
                concentration(i_stat)%may_localize=sites(j_stat)%may_localize
                concentration(i_stat)%may_reject=sites(j_stat)%may_reject
                concentration(i_stat)%category=sites(j_stat)%code

                if((.not. concentration(i_stat)%ccg_flags(2:2) .eq. '.') .and. &
                     (.not. concentration(i_stat)%may_reject)) then
                   concentration(i_stat)%may_reject = .TRUE.
                   print *,'Eventnumber',trim(concentration(i_stat)%eventnumber),' with ccg_flags ', &
                        concentration(i_stat)%ccg_flags,'has had its may_reject flag changed to TRUE'
                endif

                ! Note that delta_lat and delta_lon are initialized to zero.  They are
                ! set for particular sites when the site.moves are read in read_sites
                concentration(i_stat)%lat=concentration(i_stat)%lat + sites(j_stat)%delta_lat
                concentration(i_stat)%lon=concentration(i_stat)%lon + sites(j_stat)%delta_lon

             endif

          enddo

          if(.not. site_found) then
             print *, 'ERROR!  Station ',goUpCase(trim(concentration(i_stat)%ident)),' NOT FOUND in sites array.'
             write(gol,'("Review sites_and_weights.rc file.")');call goErr
          endif

       end do

       ! close stationfile
       close(sunit0)


       ! Check to see whether there are any "duplicate" samples.  Duplicate samples
       ! are limited to flasks only, and must be spatially and temporally co-located.
       !
       ! Colocation is defined as:
       !    within 1 hour temporally
       !    within 10m vertically
       !    within 0.05 degrees laterally (nominally, about 5km)
       !
       ! If N duplicates are detected, increase error on each by sqrt(N)
       
       i_dupe = 0
       do i_stat=1, n_fcast 

          if(.not. concentration(i_stat)%dupe .eq. 0) then
             cycle
          endif

          if(.not. concentration(i_stat)%ident(7:7) .eq. 'D') then
             cycle
          endif

          n_dupes = 0 ! number of obs in current duplicate set
          dupes = 0 ! this array keeps track of which obs are part of current duplicate set
          do j_stat = i_stat+1,n_fcast

             if(.not. concentration(j_stat)%dupe .eq. 0) then
                cycle
             endif

             if(.not. concentration(j_stat)%ident(7:7) .eq. 'D') then
                cycle
             endif

             if((abs(concentration(i_stat)%itau-concentration(j_stat)%itau) .lt. 3600) .and.  &
                (abs(concentration(i_stat)%height-concentration(j_stat)%height) .lt. 10) .and.  &
                (sqrt((concentration(i_stat)%lon-concentration(j_stat)%lon)**2 + & 
                (concentration(i_stat)%lat-concentration(j_stat)%lat)**2) .lt. 0.05)) then

                n_dupes = n_dupes + 1
                i_dupe = i_dupe + 1
                if(n_dupes .eq. 1) then
                   dupes(n_dupes) = i_stat
                   concentration(i_stat)%dupe = i_dupe
                   n_dupes = n_dupes + 1
                endif 
                concentration(j_stat)%dupe = i_dupe
                dupes(n_dupes) = j_stat
            endif
          end do

          do j_stat = 1,n_dupes
             concentration(dupes(j_stat))%error = sqrt(1.0*n_dupes)*concentration(dupes(j_stat))%error
          end do
       end do

       do i_stat = 1,n_fcast

          print('(a24,1x,f8.2,f8.2,f8.1,1x,6i4,x,f9.3,x,a3,x,a30,x,f6.2,x,i3)'), &
               concentration(i_stat)%ident, &
               concentration(i_stat)%lat, &
               concentration(i_stat)%lon, &
               concentration(i_stat)%height, &
               concentration(i_stat)%idate, &
               concentration(i_stat)%obs, &
               concentration(i_stat)%ccg_flags, &
               concentration(i_stat)%eventnumber, &
               concentration(i_stat)%error, &
               concentration(i_stat)%dupe

       end do

    endif ! if myid is root

    ! determine assignment of station to region 
    ! (with highest refinement factor)

#ifdef MPI    
    do i_stat=1, n_fcast 
       call MPI_BCAST(concentration(i_stat)%ident ,24,        MPI_CHARACTER, &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%lat   ,1,        MY_REAL      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%lon   ,1,        MY_REAL      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%height,1,        MY_REAL      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%idate,6,        MPI_INTEGER      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%eventnumber,30,        MPI_CHARACTER      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%ccg_flags,3,        MPI_CHARACTER      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%flag,1,        MY_REAL      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%itau,1,        MPI_INTEGER      , &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%obs,1,        MY_REAL,   &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%is    ,1,        MPI_INTEGER,   &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%js    ,1,        MPI_INTEGER,   &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%lsmin ,1,        MPI_INTEGER,   &
            root ,MPI_COMM_WORLD,ierr)  
       call MPI_BCAST(concentration(i_stat)%lsmax ,1,        MPI_INTEGER,   &
            root ,MPI_COMM_WORLD,ierr)  
    end do
#endif
    do i_stat=1, n_fcast

       ! assume global region as default for average mixing ratios
       concentration(i_stat)%region = 1
       concentration(i_stat)%is=1
       concentration(i_stat)%js=1
       concentration(i_stat)%n_accumulate = 0 ! no model values addded yet
       concentration(i_stat)%model = 0.0 ! no model values addded yet

       do region=1, nregions

          dyr = dy/yref(region)
          dxr = dx/xref(region)

          if ( (concentration(i_stat)%lon .gt. xbeg(region) .and. &
               concentration(i_stat)%lon .lt. xend(region)) .and.      &
               (concentration(i_stat)%lat .gt. ybeg(region) .and. &
               concentration(i_stat)%lat.lt.yend(region) ) ) then
             !=====================
             ! station is in region
             !=====================

             ! determine region with hightes refinement factor
!             if (xref(region) > xref(concentration(i_stat)%region)) then
                concentration(i_stat)%region = region
                concentration(i_stat)%is = &
                     int((concentration(i_stat)%lon-float(xbeg(region)))/dxr + 0.99999) 
                concentration(i_stat)%js = &
                     int((concentration(i_stat)%lat-float(ybeg(region)))/dyr + 0.99999)
 !            end if
          end if
       end do
    end do

    call Par_Barrier

    if(myid == root) then
       do i_stat=1, n_fcast 
          print '(a24,1x,f8.2,f8.2,f8.1,1x,3i4,1x,f9.3)', &
               concentration(i_stat)%ident, &
               concentration(i_stat)%lat, &
               concentration(i_stat)%lon, &
               concentration(i_stat)%height, &
               concentration(i_stat)%region, &
               concentration(i_stat)%is, &
               concentration(i_stat)%js, &
               concentration(i_stat)%obs
       end do
    end if

    if(allocated(dupes)) deallocate(dupes)  

    return

    ! error handling
1000 call escape_tm( 'read_co2_obs: could not open ' // ObslistFFilename) 

  end subroutine read_co2_obs

  subroutine write_co2_pseudo
    use binas,      only : ae, twopi, grav
    use dims,       only : im, jm, lm, dx, dy, xref, yref, xbeg, ybeg, xend, yend
    use dims,       only : nregions, region_name, itaur, xcyc
    use chem_param, only : ntrace, ntracet, names, fscale
    use toolbox,    only : escape_tm
    use datetime,   only : date2tau
    use go_string,           only : goNum2str
    use emission_common,       only: enkf_is_inverse, enkf_is_forward

    implicit none

    ! local
    integer                           :: slen1 !string length
    integer                           :: slen2 !string length
    integer                           :: slen3 !string length
    character(len=200)                :: dummy
    integer                           :: fstatus 
    integer                           :: i_stat
    integer                           :: sunit0=333
    real                              :: pseudo_obs
    integer                           :: itrace

    ! start

    if ( myid == root ) then

       ! take average of accumulated concentrations, convert to ppm
       ! should have already been done; no harm if repeated.
       call evaluate_forecast 

       !=========================================
       ! open StationFile (with list of stations)
       !=========================================

       open(UNIT=sunit0, FORM='FORMATTED', &
            FILE='./pseudo_co2.list' , &
            ERR=1000,status='replace')      

       do i_stat=1, n_fcast 
          
          ! In inverse mode, concentration%model(1) has the mean concentration.  In forward mode,
          ! this is only the first of five tracers (bg, ff, ocn, bio, fires).
          if (enkf_is_inverse()) then

             pseudo_obs = concentration(i_stat)%model(1)

          endif

          if (enkf_is_forward()) then

             pseudo_obs = concentration(i_stat)%model(1)

             do itrace=2,ntracet ! loop over every component tracer (bg, FF, bio, fires, ocean, ...)
                pseudo_obs = pseudo_obs + concentration(i_stat)%model(itrace)- &
                     concentration(i_stat)%model(1) ! subtract bg
             end do

          endif

!          print *, '[write_co2_pseudo] istat: ',goNum2str(i_stat,'(i3)'), &
!               ' acc: ', goNum2str(concentration(i_stat)%n_accumulate,'(i4)'), &
!               ' obs: ', &
!               goNum2str(pseudo_obs,'(f6.2)'),pseudo_obs


          write(sunit0,  '(a24,1x,f8.2,f8.2,f8.1,1x,6i4,x,f9.3,x,a3,x,a30)') &
               concentration(i_stat)%ident, &
               concentration(i_stat)%lat, &
               concentration(i_stat)%lon, &
               concentration(i_stat)%height, &
               concentration(i_stat)%idate, &
               pseudo_obs, &
               concentration(i_stat)%ccg_flags, &
               concentration(i_stat)%eventnumber
       end do
          
       ! close stationfile
       close(sunit0)

    end if ! root

    return

    ! error handling
1000 call escape_tm( 'write_co2_pseudo: could not open pseudo_co2.list') 

  end subroutine write_co2_pseudo

  subroutine get_forecast(region,itau_f,force)
    !
    ! This subroutine samples the model at given locations in array
    ! concentrations and puts the values in the tag: forecast
    ! This is done only in a certain time window, or when the keyword /force is
    ! used. Thus, a model field can be sampled at sites instantaneously (force),
    ! or at appropriate time ( omit force)
    !
    ! Note: the values accumulate so care should be taken to clear the forecast
    ! and n_accumulate tags between functions. 
    !
    ! Extension by Andy, spring 2007: see evaluate_forecast routine.  It does
    ! the unit conversion and n_accumulate averaging.  Is supposed to be robust
    ! in face of accidentally calling the routine multiple times.

    use global_data, only: mass_dat, region_dat 
    use Meteo,       only: gph_dat
    use ParTools,    only: myid,ntracet_ar

    implicit none

    ! input/output
    integer,intent(in)              :: region
    ! idate_f: date for which output required...
    integer,intent(in)              :: itau_f
    logical,optional                :: force

    ! local
    real,dimension(:,:,:), pointer      :: m,gph
    real,dimension(:,:,:,:), pointer    :: rm, rxm, rym, rzm

    real,dimension(0:lm(region)) :: height
    integer   :: i,is,js,l,n,isn,jsn,ls,j,offsetj, lst, lstn, lmr, lsn
    real      :: flon,flat,falt,ris,rjs,dxr,dyr,wcx,wcy,rls, wcz


    ! start

    m => mass_dat(region)%m_t    !pointers to global arrays...
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
    gph => gph_dat(region)%data

    lmr=lm(region)

    fcast_loop: do n=1,n_fcast 

       !  0. Is idate equal to idate_site
       !  1. Is the site in the area?--->no, then put -1 in c
       !  2. Determine gridbox
       !  3. Use slopes to determine concentration at the site.

       if ( (concentration(n)%itau.ge.itau_f-window/2.and.&
            concentration(n)%itau.lt.itau_f+window/2).or.force ) then


          dyr = dy/yref(region)
          dxr = dx/xref(region)

          flon=concentration(n)%lon
          flat=concentration(n)%lat
          falt=concentration(n)%height

          if (concentration(n)%region == region) then
             ris = (flon-float(xbeg(region)))/dxr + 0.99999
             rjs = (flat-float(ybeg(region)))/dyr + 0.99999
             is  = int(ris)   ! i-index of grid cell in which station is located
             js  = int(rjs)   ! j-index of grid cell in which station is located
             !fraction from the center of the is-box  (-0.5---+0.5)
             ris = ris-is-0.5
             !idem js
             rjs = rjs-js-0.5

             !the neighbour for pressure interpolation
             if(ris .gt. 0) then 
                isn = is+1      
             else
                isn = is-1
             endif

             !the neighbour for y interpolation
             if(rjs .gt. 0) then 
                jsn = js+1      
             else
                jsn = js-1
             endif

             ! x- / y-weighting of grid cell in which station is located
             wcx = (1.0-abs(ris))    ! 1.0 ... 0.5
             wcy = (1.0-abs(rjs))    ! 1.0 ... 0.5
             
             !=================================================================
             ! if index of neighbour is exceeding range of region set 
             ! neighbour = current cell (i.e. no interpolation)
             ! in case of cyclic x-boundaries take corresponding cyclic i index
             !=================================================================
             if ( jsn < 1) jsn=1
             if ( jsn > jm(region) ) jsn=jm(region)  ! isn-->jsn (wouter Peters)
             if ( xcyc(region) == 0 ) then
                ! non-cyclic boundaries
                if ( isn < 1) isn=1
                if ( isn > im(region) ) isn=im(region)
             else
                ! cyclic x-boundaries
                if ( isn < 1 ) isn=im(region)
                if ( isn > im(region) ) isn=1
             end if
             
             ! interpolate the altitude to site position...
             ls = 1   !layer
             do l=0,lm(region)
                height(l) =     wcx *      wcy*  gph(is,js,l+1)  + &
                     (1.0-wcx)*      wcy*  gph(isn,js,l+1) + &
                     wcx *(1.0-wcy)*  gph(is,jsn,l+1) + &
                     (1.0-wcx)*(1.0-wcy)*  gph(isn,jsn,l+1)

                if(l+1 .le. n_gph_levs) then
                   concentration(n)%gph(l+1)=concentration(n)%gph(l+1)+height(l)
                endif

             enddo
             do l=2,lm(region)   ! selects layer , note that we start from second layer from surface
                if(height(l).gt.falt) exit
             enddo

             select case(l)
             case(0)
                if(myid==root) print*,'get_forecast: Warning..., forecast altitude ',&
                     'is below the surface height',height(0),' > ',falt,concentration(n)%ident
                ls = 1
                rls = -0.5   !surface...
             case default
                ls = l       !the site layer
                ! the off-set from the center of the layer (-0.5--->+0.5)
                ! (interpolation is in (m))
                rls = (falt-height(l-1))/(height(l)-height(l-1)) - 0.5
             end select
             !=================================
             !the neighbour for z interpolation
             !=================================
             if ( rls .gt. 0 ) then 
                lsn = ls+1     
             else
                lsn = ls-1
             end if
             ! z-weighting of grid cell in which station is located 
             wcz = (1.0-abs(rls))  !.0 ... 0.5

             !=========================================================
             ! if vertical neighbor is 0 (which does not exist)
             ! take vertical layer with l=2 for EXTRApolation to ground
             !=========================================================

             IF(lsn == 0) THEN
                lsn=2
                wcz=1.0-rls  ! 1.0 ... 1.5
             ENDIF
             IF(lsn == lmr+1) THEN
                !=========================================================
                ! if vertical neighbor is lmr+1 (which does not exist)
                ! -> no interpolation
                !=========================================================
                lsn=lmr      ! no interpolation 
                wcz=1.0
             ENDIF


             !from is,js,ls, ris,rjs,rls, determine the mixing ratio of each tracer  ...
             do j=1,ntracetloc
                ! rm-value is obtained from rm + slopes. 
                ! slope = rxm = (rm*dX/dx *deltaX/2)

                ! For EnKF forward, not sure that sum over a zero
                ! length subarray works properly.  Let's be specific
                ! anyway.
                if(enkf_is_forward()) then
                   offsetj=0
                else
                   offsetj = sum(ntracet_ar(0:myid-1))
                endif

                !             if (lstn <= lmr) then   
                !                   rmf = wcx      *wcy      *wcz         *rm(is,js,ls,j)   /m(is,js,ls)   + & 
                !                        (1.0-wcx)*wcy      *wcz         *rm(isn,js,ls,j)  /m(isn,js,ls)  + &
                !                        wcx      *(1.0-wcy)*wcz         *rm(is,jsn,ls,j)  /m(is,jsn,ls)  + &
                !                        (1.0-wcx)*(1.0-wcy)*wcz         *rm(isn,jsn,ls,j) /m(isn,jsn,lst) + &
                !                        wcx      *wcy      *(1.0-wcz)   *rm(is,js,lstn,j)  /m(is,js,lstn)  + & 
                !                        (1.0-wcx)*wcy      *(1.0-wcz)   *rm(isn,js,lstn,j) /m(isn,js,lstn) + &
                !                        wcx      *(1.0-wcy)*(1.0-wcz)   *rm(is,jsn,lstn,j) /m(is,jsn,lstn) + &
                !                        (1.0-wcx)*(1.0-wcy)*(1.0-wcz)   *rm(isn,jsn,lstn,j)/m(isn,jsn,lstn) 
                !             else  
                !PB                   do NOT apply factor wcz here (no interpolation in z direction in this case) !!!!!

                !                   rmf = wcx      *wcy          *rm(is,js,ls,n)   /m(is,js,ls)   + & 
                !                                 (1.0-wcx)*wcy          *rm(isn,js,ls,n)  /m(isn,js,ls)  + &
                !                                 wcx      *(1.0-wcy)    *rm(is,jsn,ls,n)  /m(is,jsn,ls)  + &
                !                                 (1.0-wcx)*(1.0-wcy)    *rm(isn,jsn,ls,n) /m(isn,jsn,ls) 
                !              endif

                rmf = (      rm(is,js,ls,j) + &
                     2.0*(ris*rxm(is,js,ls,j) + &
                          rjs*rym(is,js,ls,j) + &
                          rls*rzm(is,js,ls,j) ) )/m(is,js,ls)*fscale(offsetj+j) 

                !    rmf = rm(is,js,ls,j)/m(is,js,ls)*fscale(offsetj+j)

                concentration(n)%model(j)=concentration(n)%model(j)+rmf !*fscale(offsetj+j)

             enddo

             concentration(n)%n_accumulate=concentration(n)%n_accumulate+1
!!             if(myid==root) then
!                print*,'DEBUG Check ',myid, concentration(n)%ident,concentration(n)%n_accumulate,concentration(n)%model(:)/concentration(n)%n_accumulate
!             endif

          endif
       endif ! force/time

    enddo fcast_loop

    nullify(m)
    nullify(rm)
    nullify(rxm)
    nullify(rym)
    nullify(rzm)
    nullify(gph)


  end subroutine get_forecast

  subroutine evaluate_forecast

    ! convert concentrations to ppm and average over n_accumulate

    integer :: n

    do n=1,ntracetloc
       where(concentration%n_accumulate.ge.1) 
          concentration%model(n)=1.e6*concentration%model(n)/concentration%n_accumulate
       end where

    enddo

    do n=1,n_gph_levs
       where(concentration%n_accumulate.ge.1) 
          concentration%gph(n)=concentration%gph(n)/concentration%n_accumulate
       end where
    enddo

    concentration%n_accumulate = 0 

  end subroutine evaluate_forecast

end module user_output_forecast
